from casadi import *
import numpy as NP
import time

from scipy import linalg


def EKF_init(f, x, p, p_known, xdot, tf, tn,t0,Q):
    # Creates EKF integrator
    Z = vertcat([x,p])
    nx = x.shape[0]
    np = p.shape[0]
    nz = nx+np
    print "nz:", nz
    print "nx:", nx
    print "np:", np
    dfdx = f.jac(0)
    dfdp = f.jac(2)[:,0:np]
    F1 = horzcat([dfdx, dfdp])
    zeroMat = NP.zeros((np, nz))
    F = vertcat([F1, zeroMat])
    P = ssym("Pkm1", nz, nz)
    P_expr = mul(F,P) + mul(P,F.T)+Q # P_dot = FP + PF'+Q
    Z.append(flatten(P))
    Z_dot_expr = vertcat([xdot,NP.zeros(np)])
    Z_dot_expr.append(flatten(P_expr))

    Zdot = SXFunction(daeIn(x=Z, p=p_known), daeOut(ode=Z_dot_expr))
    Z_int = CVodesIntegrator(Zdot)
    Z_int.setOption('t0', 0)
    Z_int.setOption("tf", (tf-t0)/tn)
    Z_int.setOption("abstol",1e-3) # tolerance
    Z_int.setOption("reltol",1e-3) # tolerance
    Z_int.init()

    return Z_int

def EKF(Z_int, z_prev, P_prev, y_k, p_known, R, H_k, M_k):
    # Z_int: integrator for combined states x, P
    # z_prev: z_{k-1}^+
    # P_prev: P_{k-1}^+
    # y_k: current measurement
    # p_known: known variables
    # R: R matrix
    # H_k: partial derivative of h with respect to z
    # M_k: partial derivative of h with respect to v

    nz = z_prev.shape[0]

    # Return x_{k}^+, P_{k}^+
    Z_prev = vertcat([flatten(z_prev), flatten(P_prev)])

    # integrate EKF
    Z_int.input(INTEGRATOR_X0).set(Z_prev)
    Z_int.input(INTEGRATOR_P).set(p_known)
    Z_int.evaluate()
    Z_out = Z_int.output()
    z_kmin = Z_out[0:nz]
    P_kmin = Z_out[nz:nz**2+nz]
    P_kmin = reshape(P_kmin,nz,nz)
    # Measurement: p_p, p_c, q_bit
    h_k = mul(H_k,z_kmin)

    # Update state and covariance estimate
    K1 = mul(P_kmin, H_k.T)
    K2 = mul(mul(H_k, P_kmin), H_k.T)
    K3 = mul(mul(M_k, R), M_k.T)
    K_k = mul(K1, linalg.inv(K2 + K3))
    IK = (NP.identity(nz)-mul(K_k, H_k))
    z_kplus = z_kmin + mul(K_k, (y_k - h_k))

    P1  = mul(mul(IK,P_kmin), IK.T)
    P21 = mul(K_k, M_k)
    P2 = mul(mul(P21,R),P21.T)
    P_kplus = P1 + P2
    return NP.array(z_kplus), NP.array(P_kplus)

def iceConcentrationHEKF(f_dae, x, p, xdot,data,t0,tf,tn,p_known,np,ny,M_k,H_k,Q,R,ObsMatFcn):
    tic = time.clock() # Begin counting time (see last part of this code)
    
    N = data.shape[0] # Number of data samples, tn and N should be equal..
    if N!=tn:
        print "Warning!"
        print "N=",N,"and tn =",tn
    
    # Dimensions
    nx = x.shape[0]

    Z_int = EKF_init(f_dae,x,p[0:np],p[np:p.shape[0]],xdot,tf,tn,t0,Q)
    # Tuneable
    P0  = DMatrix.eye(nx+np)*1
    z0 = DMatrix.zeros(nx+np,1)
    nz = nx+np
    z0[nz-3] = 0
    z0[nz-2] = 0
    z0[nz-1] = 0

    z_kmin = z0
    P_kmin = P0
    z_EKF = NP.zeros((N,nx+np))
    obsRank = NP.zeros((N,1))
    nz = z_kmin.shape[0]
    
    for k in xrange(N):
        print "Iteration number: ", k+1
        tac = time.clock()
        x_k = data[k,:]
        y_k = mul(H_k,vertcat([x_k,DMatrix.zeros(np,1)]))

        z_kplus, P_kplus = EKF(Z_int, z_kmin, P_kmin,  y_k, p_known, R, H_k, M_k)
        z_EKF[k,:] = z_kplus.T
        
        # Update for next iteration
        z_kmin = z_kplus
        P_kmin = P_kplus
        ObsMatFcn.setInput(z_kmin[:nx],0)
        ObsMatFcn.setInput(vertcat([z_kmin[nx:],p_known]),1)
        ObsMatFcn.evaluate()
        ObsMat = ObsMatFcn.output()
        obsRank[k]=NP.linalg.matrix_rank(ObsMat)

        print "y_k-z_kmin[0:nx]", vertcat([x_k])-vertcat([z_kmin[0:nx]])
        print "z_kmin[0:nx]    ", vertcat([z_kmin[0:nx]])
        print "z_kmin[nx,nx+np]", vertcat([z_kmin[nx:nx+np]])
        print k+1,". iteration took:", time.clock()-tac, "s", "ETA:", (N-(k+1))*(time.clock()-tac), "s"
        #end for
    
    print obsRank
    NP.savetxt("HEKFestimationData.txt", z_EKF, delimiter = ',')
    print "HEKF finished, and saved"      
    print "total time:", time.clock() - tic, "s"

    #END FOR
