from casadi import *
import numpy as NP
import time

from scipy import linalg


def EKF_init(f, x, p, p_known, xdot, tf, tn,t0,Q):
    # Creates EKF integrator
    Z = vertcat([x,p])
    nx = x.shape[0]
    np = p.shape[0]
    nz = nx+np
    print "nz:", nz
    print "nx:", nx
    print "np:", np
    dfdx = f.jac(0)
    dfdp = f.jac(2)[:,0:np]
    F1 = horzcat([dfdx, dfdp])
    zeroMat = NP.zeros((np, nz))
    F = vertcat([F1, zeroMat])
    P = ssym("Pkm1", nz, nz)
    P_expr = mul(F,P) + mul(P,F.T)+Q # P_dot = FP + PF'+Q
    Z.append(flatten(P))
    Z_dot_expr = vertcat([xdot,NP.zeros(np)])
    Z_dot_expr.append(flatten(P_expr))

    Zdot = SXFunction(daeIn(x=Z, p=p_known), daeOut(ode=Z_dot_expr))
    Z_int = CVodesIntegrator(Zdot)
    Z_int.setOption('t0', 0)
    Z_int.setOption("tf", (tf-t0)/tn)
    Z_int.setOption("abstol",1e-3) # tolerance
    Z_int.setOption("reltol",1e-3) # tolerance
    Z_int.init()

    return Z_int

def EKF(Z_int, z_prev, P_prev, y_k, p_known, R, H_k, M_k):
    # Z_int: integrator for combined states x, P
    # z_prev: z_{k-1}^+
    # P_prev: P_{k-1}^+
    # y_k: current measurement
    # p_known: known variables
    # R: R matrix
    # H_k: partial derivative of h with respect to z
    # M_k: partial derivative of h with respect to v

    nz = z_prev.shape[0]

    # Return x_{k}^+, P_{k}^+
    Z_prev = vertcat([flatten(z_prev), flatten(P_prev)])

    # integrate EKF
    Z_int.input(INTEGRATOR_X0).set(Z_prev)
    Z_int.input(INTEGRATOR_P).set(p_known)
    Z_int.evaluate()
    Z_out = Z_int.output()
    z_kmin = Z_out[0:nz]
    P_kmin = Z_out[nz:nz**2+nz]
    P_kmin = reshape(P_kmin,nz,nz)
    # Measurement: p_p, p_c, q_bit
    h_k = mul(H_k,z_kmin)

    # Update state and covariance estimate
    K1 = mul(P_kmin, H_k.T)
    K2 = mul(mul(H_k, P_kmin), H_k.T)
    K3 = mul(mul(M_k, R), M_k.T)
    K_k = mul(K1, linalg.inv(K2 + K3))
    IK = (NP.identity(nz)-mul(K_k, H_k))
    z_kplus = z_kmin + mul(K_k, (y_k - h_k))

    P1  = mul(mul(IK,P_kmin), IK.T)
    P21 = mul(K_k, M_k)
    P2 = mul(mul(P21,R),P21.T)
    P_kplus = P1 + P2
    return NP.array(z_kplus), NP.array(P_kplus)

def iceConcentrationMHE(f_dae, x, p, xdot,data,t0,tf,tn,nk,p_known,np,Vj,ny,M_k,H_k,Q,R,EnableEKFarrivalCost,ObsMatFcn):
    tic = time.clock() # Begin counting time (see last part of this code)
    
    N = data.shape[0] # Number of data samples, tn and N should be equal..
    if N!=tn:
        print "Warning!"
        print "N=",N,"and tn =",tn
    
    # get model
    f = SXFunction([x,p],[xdot])
    # Dimensions
    nx = x.shape[0]
    
    
    ## HEKF begin
    Z_int = EKF_init(f_dae,x,p[0:np],p[np:p.shape[0]],xdot,tf,tn,t0,Q)
    # Tuneable
    P0  = DMatrix.eye(nx+np)*1
    z0 = DMatrix.zeros(nx+np,1)
    nz = nx+np
    z0[nz-3] = 0
    z0[nz-2] = 0
    z0[nz-1] = 0

    z_kmin = z0
    P_kmin = P0
    z_EKF = NP.zeros((N,nx+np))
    obsRank = NP.zeros((N,1))
    nz = z_kmin.shape[0]
    
    # Variables for the cost function
    xk = ssym("xk",nx,1)
    xk = vertcat([xk])
    data_k = ssym("data_k",nx,1)
    data_k = vertcat([data_k])
    zk = ssym("zk",nz,1)
    zk = vertcat([zk])
    pk = ssym("pk",np,1)
    pk = vertcat([pk])
    P_mat = ssym("P_mat",nz,nz)

    # Cost function
    m = SXFunction([xk,data_k],[mul(mul(xk.T-data_k.T,Vj),xk-data_k)]) # (xk^T - data_k^T)Vj(xk - data_k) calculating the quadratic cost.
    ArrivalCost = SXFunction([xk,pk,zk,P_mat],[mul(mul(vertcat([xk,pk]).T-zk[0:nz].T,P_mat),vertcat([xk,pk])-zk[0:nz])]) # (xk^T - z_kmin^T)P_kmin^-1(xk - z_kmin) calculating the quadratic cost.
    # Initialize functions
    f.init()
    m.init()
    ArrivalCost.init()
    
    # Legendre collocation points
    legendre_points1 = [0,0.500000]
    legendre_points2 = [0,0.211325,0.788675]
    legendre_points3 = [0,0.112702,0.500000,0.887298]
    legendre_points4 = [0,0.069432,0.330009,0.669991,0.930568]
    legendre_points5 = [0,0.046910,0.230765,0.500000,0.769235,0.953090]
    
    # Radau collocation points
    radau_points1 = [0,1.000000]
    radau_points2 = [0,0.333333,1.000000]
    radau_points3 = [0,0.155051,0.644949,1.000000]
    radau_points4 = [0,0.088588,0.409467,0.787659,1.000000]
    radau_points5 = [0,0.057104,0.276843,0.583590,0.860240,1.000000]

    # Choose collocation points
    tau_root = radau_points5
    
    # Degree of interpolating polynomial
    d = len(tau_root)-1
    
    # Total number of variables
    NX = nk*(d+1)*nx        # Collocated states
    NXF = nx                # Final states
    NV = NX+NXF+np          # Total number of variables
    
    # Size of the finite elements
    h = (tf-t0)/(N-1) # N-1 since the first measurement takes place at t0.
    
    x_opt_iterate = NP.zeros(NV)
    x_opt = NP.zeros((N-nk,nx))
    p_opt = NP.zeros((N-nk,np))
 
 
    
    # Coefficients of the collocation equation
    C = NP.zeros((d+1,d+1))

    # Coefficients of the continuity equation
    D = NP.zeros(d+1)

    # Dimensionless time inside one control interval
    tau = ssym("tau")

    # For all collocation points
    for j in range(d+1): #range(d+1) = 0,1,...,d
        # Construct Lagrange polynomials to get the polynomial basis at the collocation point
        L = 1
        for i in range(d+1):
            if i != j:
                L *= (tau-tau_root[i])/(tau_root[j]-tau_root[i])
                #END IF
            #END FOR
        lfcn = SXFunction([tau],[L]) #l_j(tau)
        lfcn.init()
        
        # Evaluate the polynomial at the final time to get the coefficients of the continuity equation
        lfcn.setInput(1.0)
        lfcn.evaluate()
        D[j] = lfcn.output() #Dj = l_j(1)
        
        # Evaluate the time derivative of the polynomial at all collocation points to get the coefficients of the continuity equation
        for i in range(d+1):
            lfcn.setInput(tau_root[i])
            lfcn.setFwdSeed(1.0)
            lfcn.evaluate(1,0)
            C[j,i] = lfcn.fwdSens() #Cji = l_j'(tau_i)
            #END FOR
        #END FOR

    

    # NLP variable vector
    V = MX("V",NV)
    # NLP parameter vector
    data_input = MX("data_input",(nk+1)*nx+nz+nz*nz) #Measured data, KF-data, P-matrix
    
    init_states = NP.zeros(nx)
    
    # State bounds and initial guess
    x_min =  NP.zeros(nx)
    x_max =  NP.zeros(nx)+1
    xi_min = NP.zeros(nx) #Boundary for init states
    xi_max = NP.zeros(nx)+1
    xf_min = NP.zeros(nx) #Boundary for end states
    xf_max = NP.zeros(nx)+1
    x_init = init_states # Init states..
    
    # Parameter bounds and initial guess
    p_min = NP.zeros(np)
    p_max = NP.zeros(np)+1
    if np == 0:
        print "No parameters to estimate"
    elif np+len(p_known) == 3: # 1D
        p_min[0] = -3 # Minimum ice drift velocity
        p_max[0] = 3 # Maximum ice drift velocity
    else: # 2D
        p_min[0] = -3
        p_min[1] = -3
        p_max[0] = 3
        p_max[1] = 3
        
    p0 = NP.zeros(np)
  
    # All variables with bounds and initial guess
    vars_lb = NP.zeros(NV)
    vars_ub = NP.zeros(NV)
    vars_init = NP.zeros(NV)
    offset = 0

    # Get collocated states and parametrized control
    X = NP.resize(NP.array([],dtype=MX),(nk+1,d+1))
    for k in range(nk):  
        # Collocated states
        for j in range(d+1):
            # Get the expression for the state vector
            X[k,j] = V[offset:offset+nx]
            
            # Add the initial condition
            vars_init[offset:offset+nx] = x_init
            
            # Add bounds
            if k==0 and j==0:
                vars_lb[offset:offset+nx] = xi_min
                vars_ub[offset:offset+nx] = xi_max
            else:
                vars_lb[offset:offset+nx] = x_min
                vars_ub[offset:offset+nx] = x_max
                #END IF
            offset += nx
            #END FOR

    # State at end time
    X[nk,0] = V[offset:offset+nx]
    vars_lb[offset:offset+nx] = xf_min
    vars_ub[offset:offset+nx] = xf_max
    vars_init[offset:offset+nx] = x_init
    offset += nx

    # Parameters
    P = V[offset:offset+np]
    
    vars_lb[offset:offset+np] = p_min
    vars_ub[offset:offset+np] = p_max
    vars_init[offset:offset+np] = p0 
    offset += np
    
    # Constraint function for the NLP
    g = []
    lbg = []
    ubg = []

    # For all finite elements
    for k in range(nk):
        # For all collocation points
        for j in range(1,d+1):
            # Get an expression for the state derivative at the collocation point
            xp_jk = 0
            for i in range (d+1):
                xp_jk += C[i,j]*X[k,i]
                #END FOR
            # Add collocation equations to the NLP
            if np>0: # If velocity is the only unknown parameter
                [fk] = f.call([X[k,j], vertcat([P,p_known])])
            else: # If all the parameters are unknown
                [fk] = f.call([X[k,j], vertcat([p_known])])
                    
            g.append(h*fk - xp_jk)
            lbg.append(NP.zeros(nx)) # equality constraints
            ubg.append(NP.zeros(nx)) # equality constraints
            #END FOR
          
        # Get an expression for the state at the end of the finite element
        xf_k = 0
        for i in range(d+1):
            xf_k += D[i]*X[k,i]
            #END FOR
          
        # Add continuity equation to NLP
        g.append(X[k+1,0] - xf_k)
        lbg.append(NP.zeros(nx))
        ubg.append(NP.zeros(nx))
        #END FOR
  
  
    # Concatenate constraints
    g = vertcat(g)

    # Objective function of the NLP
    J = 0
    [f2] = ArrivalCost.call([X[0,0],P,data_input[(nk)*nx+nx:(nk)*nx+nx+nz],reshape(data_input[-(nz*nz):],(nz,nz))])
    J += f2 *EnableEKFarrivalCost
    for i in range(nk+1): # For all discretized states
        [f1] = m.call([X[i,0],data_input[i*nx:i*nx+nx]]) # Calculate the quadratic cost
        J += f1

    ## ----
    ## SOLVE THE NLP
    ## ----
    
    # Allocate an NLP solver
    nlp = MXFunction(nlpIn(x=V,p=data_input),nlpOut(f=J,g=g))
    solver = IpoptSolver(nlp)
    # Set options
    solver.setOption("linear_solver", "ma27")
#    solver.setOption("suppress_all_output","yes")
    # initialize the solver
    print "Trying to initialize the solver.."
    solver.init()
    print "Initialized"

    ## Moving horizon over data set
    for k_mhe in xrange(N-nk):
        print "Iteration number: ", k_mhe+1
        
        tac = time.clock()
        x_k = data[k_mhe:k_mhe+nk+1,:]
        y_k = mul(H_k,vertcat([x_k[0,:],DMatrix.zeros(np,1)]))

        z_kplus, P_kplus = EKF(Z_int, z_kmin, P_kmin,  y_k, p_known, R, H_k, M_k) #TODO control that data_k is correct
        z_EKF[k_mhe,:] = z_kplus.T
        
        # Update for next iteration
        z_kmin = z_kplus
        P_kmin = P_kplus 
        input_parameters = vertcat([flatten(x_k),flatten(z_kmin),flatten(linalg.inv(P_kplus))])
        
        # Bounds on x
        solver.setInput(vars_lb,"lbx")
        solver.setInput(vars_ub,"ubx")
        
        # Bounds on g
        solver.setInput(NP.concatenate(lbg),"lbg")
        solver.setInput(NP.concatenate(ubg),"ubg")
        
        # Known parameters for the solver
        solver.setInput(input_parameters,"p")

        # Decision variables initial guess
        solver.setInput(x_opt_iterate,"x0")           
        print "Ready to solve"
        # Solve the problem
        solver.solve()
        
        # Print the optimal cost
        print "optimal cost: ", float(solver.output("f"))
        
        # Retrieve the solution
        x_opt_iterate = NP.array(solver.output("x"))

            
        for i in xrange(nx):
            x_opt[k_mhe,i]=x_opt_iterate[-(nx+np-i)]
            #end for
        for i in xrange(np):
            p_opt[k_mhe,i]=x_opt_iterate[-(np-i)]
            #end for
        pk_opt = vertcat([p_opt[k_mhe,:].T])
        
        # Calculate the linearized observability
        ObsMatFcn.setInput(x_opt_iterate[:nx],0)
        ObsMatFcn.setInput(vertcat([pk_opt,p_known]),1)
        ObsMatFcn.evaluate()
        ObsMat = ObsMatFcn.output()
        obsRank[k_mhe]=NP.linalg.matrix_rank(ObsMat)
        
        print "p_opt:", p_opt[k_mhe,:]
        print "y_k-z_kmin[0:nx]", vertcat([x_k[0,:]])-vertcat([z_kmin[0:nx]])
        print "z_kmin[0:nx]    ", vertcat([z_kmin[0:nx]])
        print "z_kmin[nx,nx+np]", vertcat([z_kmin[nx:nx+np]])
        print k_mhe+1,". iteration took:", time.clock()-tac, "s", "ETA:", (N-nk-(k_mhe+1))*(time.clock()-tac), "s"
        #end for
    

    # Run the last iterations for the Kalman filter and observability calculations
    for i in xrange(nk):
        ObsMatFcn.setInput(x_opt_iterate[nx*(i+1):nx*(i+1)+nx],0)
        ObsMatFcn.setInput(vertcat([pk_opt,p_known]),1)
        ObsMatFcn.evaluate()
        ObsMat = ObsMatFcn.output()
        obsRank[k_mhe+(i+1)]=NP.linalg.matrix_rank(ObsMat)
        
        x_k = data[k_mhe+(i+1),:]
        y_k = mul(H_k,vertcat([x_k,DMatrix.zeros(np,1)]))

        z_kplus, P_kplus = EKF(Z_int, z_kmin, P_kmin,  y_k, p_known, R, H_k, M_k)
        z_EKF[k_mhe+(i+1),:] = z_kplus.T
        
        # Update for next iteration
        z_kmin = z_kplus
        P_kmin = P_kplus
        #End for
    print obsRank       
    NP.savetxt("MHEdata.txt", x_opt, delimiter =',' )
    NP.savetxt("MHE_param_data.txt", p_opt, delimiter = ',')
    NP.savetxt("HEKF_estimation_data.txt", z_EKF, delimiter = ',')
    print "MHE finished, and saved"
    
    k = k_mhe
    if np == 3:
        print "u:", p_opt[k,0]
        print "c_left:", p_opt[k,1]
        print "c_right", p_opt[k,2]
    elif np == 2:
        print "u:", p_opt[k,0]
        print "c_left:", p_opt[k,1]
    elif np == 1:
        print "u:", p_opt[k,0]
       
    print "total time:", time.clock() - tic, "s"

    #END FOR
