## The advection equation discretized and implemented by Joakim Haugen
## Modified in order to get a uniform u

from numpy import *
import numpy
from casadi import *

def iceConcentrationModel(x_R,x_L,y_R,y_L,r,c):
    # Algorithm parameters
    c_L = 1.0/4
    c_R = 1.0/4
    c_C = 1.0/2
    eps = 1e-6
    
    Dx = abs((x_R-x_L)/r)
    Dy = abs((y_R-y_L)/c)
    
    def xj(j):
        return x_L+j*Dx
    def yk(k):
        return y_L+j*Dy
    
    c_bar = ssym("c",r,c) # System states
    con_L = ssym("c_L",1,c) # Concentration on the left boundary
    con_R = ssym("c_R",1,c) # Concentration on the right boundary
    con_T = ssym("c_T",r,1) # Concentration on the top boundary
    con_B = ssym("c_B",r,1) # Concentration on the bottom boundary
    
    u_interior = ssym("u",1,1)
    u_L = ssym("u_L",1,c) # ice drift velocity on the left boundary
    u_R = ssym("u_R",1,c) # ice drift velocity on the right boundary
    u_T = ssym("u_T",r,1) # ice drift velocity on the top boundary
    u_B = ssym("u_B",r,1) # ice drift velocity on the bottom boundary

    v_interior = ssym("v",1,1)
    v_L = ssym("v_L",1,c) # ice drift velocity on the left boundary
    v_R = ssym("v_R",1,c) # ice drift velocity on the right boundary
    v_T = ssym("v_T",r,1) # ice drift velocity on the top boundary
    v_B = ssym("v_B",r,1) # ice drift velocity on the bottom boundary
    
    c_field = SXMatrix(r+4,c+4)
    c_field[2:r+2,2:c+2] = c_bar
    c_field[0,2:c+2] = con_L
    c_field[1,2:c+2] = con_L
    c_field[2:r+2,0] = con_B
    c_field[2:r+2,1] = con_B
    c_field[r+2,2:c+2] = con_R
    c_field[r+3,2:c+2] = con_R
    c_field[2:r+2,c+2] = con_T
    c_field[2:r+2,c+3] = con_T

    u_field = SXMatrix(r+4,c+4)
    v_field = SXMatrix(r+4,c+4)
    for i in range(r+4):
        for j in range(c+4):
            u_field[i,j] = u_interior
            v_field[i,j] = v_interior
    
    def stencil(j,k,d):
        sten_c = SXMatrix(3,3)
        sten_c[1,1] = c_field[j+1,k+1]
        if d==0:
            sten_c[0,1] = c_field[j-1+1,k+1]
            sten_c[2,1] = c_field[j+1+1,k+1]
        else:
            sten_c[1,0] = c_field[j+1,k-1+1]
            sten_c[1,2] = c_field[j+1,k+1+1]
        return sten_c
    
    def sten_uv(j,k,d):
    
        if d==0:
            sten_u = SXMatrix(3,3)
            sten_u[1,1] = u_field[j+1,k+1]
            sten_u[0,1] = u_field[j-1+1,k+1]
            sten_u[2,1] = u_field[j+1+1,k+1]
            sten_u[1,0] = u_field[j+1,k-1+1]
            sten_u[1,2] = u_field[j+1,k+1+1]
            return sten_u
        else:
            sten_v = SXMatrix(3,3)
            sten_v[1,1] = v_field[j+1,k+1]
            sten_v[0,1] = v_field[j-1+1,k+1]
            sten_v[2,1] = v_field[j+1+1,k+1]
            sten_v[1,0] = v_field[j+1,k-1+1]
            sten_v[1,2] = v_field[j+1,k+1+1]
            return sten_v
    
    def IS_L(j,k,d):
        sten = stencil(j,k,d)
        if d==0:
            return (sten[1,1] - sten[0,1])**2
        else:
            return (sten[1,1] - sten[1,0])**2
    
    def IS_R(j,k,d):
        sten = stencil(j,k,d)
        if d==0:
            return (sten[2,1]-sten[1,1])**2
        else:
            return (sten[1,2]-sten[1,1])**2
    
    def IS_C(j,k,d):
        sten = stencil(j,k,d)
        if d==0:
            return (13.0*(sten[2,1] - 2*sten[1,1] + sten[0,1])**2)/3.0\
            + ((sten[2,1]-sten[0,1])**2)/4.0
        else:
            return (13.0*(sten[1,2] - 2*sten[1,1] + sten[1,0])**2)/3.0\
            + ((sten[1,2]-sten[1,0])**2)/4.0
    
    def alpha_L(j,k,d):
        if d==0:
            return c_L/(eps + IS_L(j,k,0))**2
        else:
            return c_L/(eps + IS_L(j,k,1))**2
    
    def alpha_R(j,k,d):
        if d==0:
            return c_R/(eps + IS_R(j,k,0))**2
        else:
            return c_R/(eps + IS_R(j,k,1))**2
    
    def alpha_C(j,k,d):
        if d==0:
            return c_C/(eps + IS_C(j,k,0))**2
        else:
            return c_C/(eps + IS_C(j,k,1))**2
    
    def sum_alpha(j,k,d):
        aL = alpha_L(j,k,d)
        aR = alpha_R(j,k,d)
        aC = alpha_C(j,k,d)
        return aL+aR+aC
    
    def w_L(j,k,d):
        return alpha_L(j,k,d)/sum_alpha(j,k,d)
    
    def w_R(j,k,d):
        return alpha_R(j,k,d)/sum_alpha(j,k,d)
    
    def w_C(j,k,d):
        return alpha_C(j,k,d)/sum_alpha(j,k,d)
    
    def PRk(x,j,k):
        sten = stencil(j,k,0)
        return sten[1,1] +(sten[2,1]-sten[1,1])*(x-xj(j))/Dx
    def PRj(y,j,k):
        sten = stencil(j,k,1)
        return sten[1,1] +(sten[1,2]-sten[1,1])*(y-yk(k))/Dy
    def PLk(x,j,k):
        sten = stencil(j,k,0)
        return sten[1,1] + (sten[1,1]-sten[0,1])*(x-xj(j))/Dx
    def PLj(y,j,k):
        sten = stencil(j,k,1)
        return sten[1,1] + (sten[1,1]-sten[1,0])*(y-yk(k))/Dy
    def PCk(x,j,k):
        sten = stencil(j,k,0)
        Pc = sten[1,1] - (sten[2,1] - 2*sten[1,1] + sten[0,1])/12.0 \
        -(sten[1,2]-2*sten[1,1]+sten[1,0])/12.0 \
        +(sten[2,1]-sten[0,1])*(x-xj(j))/(2*Dx) \
        + (sten[2,1]-2*sten[1,1]+sten[0,1])*(x-xj(j))**2/(Dx**2)
        return Pc
    def PCj(y,j,k):
        sten = stencil(j,k,1)
        Pc = sten[1,1] -(sten[1,2]-2*sten[1,1]+sten[1,0])/12.0 \
        - (sten[2,1] - 2*sten[1,1] + sten[0,1])/12.0 \
        +(sten[1,2]-sten[1,0])*(y-yk(k))/(2*Dy) \
        + (sten[1,2]-2*sten[1,1]+sten[1,0])*(y-yk(k))**2/(Dy**2)
        return Pc
    
    def cplus(j,k,d):
        if d==0:
            wL = w_L(j+1,k,d)
            wR = w_R(j+1,k,d)
            wC = w_C(j+1,k,d)
            return wL*PLk(xj(j+0.5),j+1,k)+wR*PRk(xj(j+0.5),j+1,k) + wC*PCk(xj(j+0.5),j+1,k)
        else:
            wL = w_L(j,k+1,d)
            wR = w_R(j,k+1,d)
            wC = w_C(j,k+1,d)
            return wL*PLj(yk(k+0.5),j,k+1)+wR*PRj(yk(k+0.5),j,k+1) + wC*PCj(yk(k+0.5),j,k+1)
    
    def cminus(j,k,d):
        wL = w_L(j,k,d)
        wR = w_R(j,k,d)
        wC = w_C(j,k,d)
    
        if d==0:
            return wL*PLk(xj(j+0.5),j,k)+wR*PRk(xj(j+0.5),j,k) + wC*PCk(xj(j+0.5),j,k)
        else:
            return wL*PLj(yk(k+0.5),j,k)+wR*PRj(yk(k+0.5),j,k) + wC*PCj(yk(k+0.5),j,k)
    
    def ax(j,k):
        sten = sten_uv(j,k,0)
        return fabs(sten[1,1])
    def ay(j,k):
        sten = sten_uv(j,k,1)
        return fabs(sten[1,1])
    def fx(j,k,c):
        sten = sten_uv(j,k,0)
        return sten[1,1]*c
    def gy(j,k,c):
        sten = sten_uv(j,k,1)
        return sten[1,1]*c
    
    def Hx(j,k):
        return (fx(j,k,cplus(j,k,0)) + fx(j,k,cminus(j,k,0)))/2.0 \
         - ax(j,k)*(cplus(j,k,0)-cminus(j,k,0))/2.0
    
    def Hy(j,k):
        return (gy(j,k,cplus(j,k,1)) + gy(j,k,cminus(j,k,1)))/2.0 \
         - ay(j,k)*(cplus(j,k,1)-cminus(j,k,1))/2.0
    
    dot_c = SXMatrix(r,c)
    
    for j in range(1,r+1):
        for k in range(1,c+1):
            dot_c[j-1,k-1] = - (Hx(j,k)-Hx(j-1,k))/Dx -(Hy(j,k)-Hy(j,k-1))/Dy


    dot_flat = casadi.reshape(dot_c,r*c,1)
    c_flat = casadi.reshape(c_bar,r*c,1)
    c_bound = vertcat([con_L.T,con_R.T,con_T,con_B])
    p = vertcat([u_interior,v_interior,c_bound])
    
    f_rhs = SXFunction(daeIn(x=c_flat,p=p),daeOut(ode=dot_flat))
    return f_rhs, c_flat, p, dot_flat

